unit uMidi;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, mmsystem;

type
  TMidiDevice = class;

  TMidiChannel = class(TComponent)
  private
    FChannel: Integer;
    FDevice: TMidiDevice;
    FPrevVel: ShortInt;
    class function MidiMsg(ch, st, d1, d2: Byte): Cardinal;
  public
    property Channel: Integer read FChannel;
    procedure SetVel(vel: Byte);
    procedure NoteOn(note: Byte; vel: Byte = 255);
    procedure NoteOff(note: Byte);
    procedure NoteOnDelayed(delay: Integer; note: Byte; vel: Byte = 255);
    procedure NoteOffDelayed(delay: Integer; note: Byte);
    procedure AllSoundOff;
    procedure PChange(p: Byte);
    constructor Create(AOwner: TComponent); override;
  end;

  TMidiDevice = class(TComponent)
  private
    FThreadEnabled: Boolean;
    FCriticalSection: TRTLCriticalSection;
    FThread: TThread;
    FDevice: HMIDIOUT;
    FDeviceList: TStringList;
    FDeviceID: Integer;
    FChannels: array of TMidiChannel;
    procedure Lock;
    procedure Unlock;
    function GetChannels(i: Integer): TMidiChannel;
    procedure OutShortMsg(a: Cardinal);
    function GetThreadEnabled: Boolean;
    procedure SetThreadEnabled(const Value: Boolean);
    function GetDeviceList: TStrings;
    function GetChannelCount: Integer;
  public
    property DeviceList: TStrings read GetDeviceList;
    property DeviceID: Integer read FDeviceID write FDeviceID;
    property Channels[i: Integer]: TMidiChannel read GetChannels;
    property ChannelCount: Integer read GetChannelCount;
    property ThreadEnabled: Boolean read GetThreadEnabled write SetThreadEnabled;
    procedure QueueShortMsg(delay: Integer; msg: Cardinal);
    procedure AllSoundOff;
    function Open: Boolean;
    procedure Close;
    procedure UpdateDeviceList;
    constructor Create(AOwner: TComponent; AChannelCount: Integer = 16); reintroduce;
    destructor Destroy; override;
  end;

implementation

type
  PMidiEvent = ^TMidiEvent;
  TMidiEvent = record
    delay: Integer;
    msg: Cardinal;
  end;

  TMidiThread = class(TThread)
  private
    FDevice: TMidiDevice;
    FEventList: TThreadList;
  public
    constructor Create;
    destructor Destroy; override;
  protected
    procedure Execute; override;
  end;

{ TMidiDevice }

procedure TMidiDevice.AllSoundOff;
var
  i: Integer;
begin
  for i := 0 to High(FChannels) do
    FChannels[i].AllSoundOff;
end;

procedure TMidiDevice.Close;
begin
  Lock;
  try
    if FDevice = 0 then
      Exit;
    midiOutClose(FDevice);
    FDevice := 0;
  finally
    Unlock;
  end;
end;

constructor TMidiDevice.Create(AOwner: TComponent; AChannelCount: Integer);
var
  i: Integer;
begin
  inherited Create(AOwner);
  InitializeCriticalSection(FCriticalSection);
  FDeviceList := TStringList.Create;
  UpdateDeviceList;
  SetLength(FChannels, AChannelCount);
  for i := 0 to High(FChannels) do
  begin
    FChannels[i] := TMidiChannel.Create(Self);
    FChannels[i].FChannel := i;
    FChannels[i].FDevice := Self;
  end;
  FThreadEnabled := True;
  FThread := TMidiThread.Create;
  TMidiThread(FThread).FDevice := Self;
end;

destructor TMidiDevice.Destroy;
begin
  FThread.Terminate;
  FThread.WaitFor;
  FThread.Free;
  Close;
  FDeviceList.Free;
  DeleteCriticalSection(FCriticalSection);
  inherited;
end;

function TMidiDevice.GetChannelCount: Integer;
begin
  Result := Length(FChannels);
end;

function TMidiDevice.GetChannels(i: Integer): TMidiChannel;
begin
  Result := FChannels[i];
end;

function TMidiDevice.GetDeviceList: TStrings;
begin
  Result := FDeviceList;
end;

function TMidiDevice.GetThreadEnabled: Boolean;
begin
  Lock;
  try
    Result := FThreadEnabled;
  finally
    Unlock;
  end;
end;

procedure TMidiDevice.Lock;
begin
  EnterCriticalSection(FCriticalSection);
end;

function TMidiDevice.Open: Boolean;
begin
  Close;
  Lock;
  try
    Result := (midiOutOpen(@FDevice, FDeviceID, 0, 0, 0) = MMSYSERR_NOERROR);
    if not Result then
      FDevice := 0;
  finally
    Unlock;
  end;
end;

procedure TMidiDevice.OutShortMsg(a: Cardinal);
begin
  Lock;
  try
    if FDevice <> 0 then
      midiOutShortMsg(FDevice, a);
  finally
    Unlock;
  end;
end;

procedure TMidiDevice.QueueShortMsg(delay: Integer; msg: Cardinal);
var
  p: PMidiEvent;
begin
  New(p);
  p^.delay := delay;
  p^.msg := msg;
  TMidiThread(FThread).FEventList.Add(p);
end;

procedure TMidiDevice.SetThreadEnabled(const Value: Boolean);
begin
  Lock;
  try
    FThreadEnabled := Value;
  finally
    Unlock;
  end;
end;

procedure TMidiDevice.Unlock;
begin
  LeaveCriticalSection(FCriticalSection);
end;

procedure TMidiDevice.UpdateDeviceList;
var
  i, c: Integer;
  caps: TMidiOutCaps;
begin
  FDeviceList.Clear;
  c := midiOutGetNumDevs;
  for i := 0 to c - 1 do
  begin
    midiOutGetDevCaps(i, @caps, Sizeof(caps));
    FDeviceList.Add(caps.szPname);
  end;
end;

{ TMidiChannel }

procedure TMidiChannel.AllSoundOff;
var
  i: Integer;
begin
  with TMidiThread(FDevice.FThread).FEventList.LockList do
  begin
    for i := 0 to Count - 1 do
      if TMidiEvent(Items[i]^).msg and 15 = Cardinal(FChannel) then
        Items[i] := nil;
    Pack;
    TMidiThread(FDevice.FThread).FEventList.UnlockList;
  end;
  FDevice.QueueShortMsg(0, MidiMsg(FChannel, $B, $78, 0));
end;

constructor TMidiChannel.Create(AOwner: TComponent);
begin
  inherited;
  FPrevVel := 127;
end;

class function TMidiChannel.MidiMsg(ch, st, d1, d2: Byte): Cardinal;
begin
  Result := ch + st shl 4 + d1 shl 8 + d2 shl 16;
end;

procedure TMidiChannel.NoteOff(note: Byte);
begin
  NoteOffDelayed(0, note);
end;

procedure TMidiChannel.NoteOffDelayed(delay: Integer; note: Byte);
begin
  FDevice.QueueShortMsg(delay, MidiMsg(FChannel, $9, note, 0));
end;

procedure TMidiChannel.NoteOn(note, vel: Byte);
begin
  NoteOnDelayed(0, note, vel);
end;

procedure TMidiChannel.NoteOnDelayed(delay: Integer; note, vel: Byte);
begin
  if vel = 255 then
    vel := FPrevVel;
  FDevice.QueueShortMsg(delay, MidiMsg(FChannel, $9, note, vel));
  FPrevVel := vel;
end;

procedure TMidiChannel.PChange(p: Byte);
begin
  FDevice.QueueShortMsg(0, MidiMsg(FChannel, $C, p, 0));
end;

procedure TMidiChannel.SetVel(vel: Byte);
begin
  FPrevVel := vel;
end;

{ TMidiThread }

constructor TMidiThread.Create;
begin
  inherited Create(True);
  timeBeginPeriod(1);
  FreeOnTerminate := False;
  FEventList := TThreadList.Create;
  Resume;
end;

destructor TMidiThread.Destroy;
var
  i: Integer;
  list: TList;
  p: PMidiEvent;
begin
  list := FEventList.LockList;
  try
    for i := list.Count - 1 downto 0 do
    begin
      p := list[i];
      Dispose(p);
    end;
  finally
    FEventList.UnlockList;
  end;
  FEventList.Free;
  timeEndPeriod(1);
  inherited;
end;

procedure TMidiThread.Execute;
var
  i, j: Integer;
  tm, dt, prevTime: Cardinal;
  justStarted: Boolean;
  list: TList;
  p: PMidiEvent;
begin
  inherited;

  justStarted := True;
  prevTime := 0;

  while not Terminated do
  begin
    Sleep(1);

    if not FDevice.ThreadEnabled then
    begin
      justStarted := True;
      Continue;
    end;

    tm := timeGetTime;
    if justStarted then
    begin
      dt := 0;
      justStarted := False;
    end
    else
      dt := tm - prevTime;
    prevTime := tm;

    if dt > 1000 then
      dt := 1000;
    //dtُȒlɂȂƃXgbv邩

    list := FEventList.LockList;
    try
      for j := 1 to dt do //͎ԏۏႷ邽
      begin
        for i := 0 to list.Count - 1 do
        begin
          p := list[i];

          Dec(p^.delay);
          if p^.delay <= 0 then
          begin
            FDevice.OutShortMsg(p^.msg);
            Dispose(p);
            list[i] := nil;
          end;
        end;
        list.Pack;
      end;
    finally
      FEventList.UnlockList;
    end;
  end;
end;

end.
